Object subclass: #AbstractProto
	instanceVariableNames: 'organization '
	classVariableNames: 'ReturnFromSignal UnknownGlobalVariable UnknownMethod '
	poolDictionaries: ''
	category: 'Prototypes'!
AbstractProto comment:
'This class is the abstracyt class for all prototype classes.
It represents the most simple version of prototypes.
No sharing (no parent link and no delegation), one method dictionary (instance variable ''methodes''), a clone operation.
Method simply returning a value allow users to similate object variables (we should no longer speak of instance variables).
Methodes are invoked via message sending.
A variable within a program is interpreted as a global variable.

In this class, the categories designated as ''private'' own the methods for the implementation and the categories designated as ''primitive'' own method which are the primitive of the associated prototype language, e.g. the method ''method:'' to define new methods on prototypes.

Method defined on prototypes are represented as abstract syntax trees created by the smalltalk compiler.
The default node builder is ''ProgramNodeBuilder'' which defines standrad smalltalk program nodes.
The evaluators for the prototype languages are defined as methods named ''eval'' and ''apply'' on the nodes classes
(e.g. SimpleMessageNode).

To launch the prototype language associated with a class, see the class-method ''example''.

See also the hierarchy.
'!


!AbstractProto methodsFor: 'adding primitive (private)'!

addPrimitive: aCompiledMethod named: sel
	self subclassResponsibility!

addPrimitiveNamed: sel
	| aClass |
	aClass := self class SmalltalkMethodOwner: sel.
	self addPrimitive: (aClass compiledMethodAt: sel) named: sel.
	^sel! !

!AbstractProto methodsFor: 'initializing (private)'!

initialize
	"The organization of methods within a class appears to be the same than the
	organization of methods within prototypes, we can thus use ClassOrganizer"
	organization := ClassOrganizer new.! !

!AbstractProto methodsFor: 'deep cloning (primitive)'!

deepClone
	^self deepCopy! !

!AbstractProto methodsFor: 'browser Interface'!

browse
	"send the message browse to a proto"
	ProtoBrowserView openOn: self type: self class defaultBrowser!

fileOut: fileStream name: externalName
	(organization categories) do: 
		[:protocol | 
		self fileOutProtocol: protocol file: fileStream name: externalName.]!

fileOutMessage: selector file: fileStream name: externalName protocol: protocol
	fileStream cr. fileStream cr.
	self printMessageChunk: selector file: fileStream name: externalName protocol: protocol.!

fileOutProtocol: protocol file: fileStream name: externalName
	fileStream cr. fileStream cr.
	fileStream nextChunkPut: '"methods for ' , protocol , '"'.
	(organization listAtCategoryNamed: protocol) do: 
		[:selector | 
		self fileOutMessage: selector file: fileStream name: externalName protocol: protocol.]!

name
	"up to now, prototypes are unnamed"
	^'unnamed proto'!

organization
	^organization!

printMessageChunk: selector file: fileStream name: externalName protocol: protocol 
	Cursor write showWhile: [fileStream nextChunkPut: (externalName , '\  method: ' , '''' , (self compiledMethodAt: selector) decompileString , '''\  classified: ''' , protocol , '''') withCRs]!

whichCategoryIncludesSelector: aSelector 
	"Answer the category of the argument, aSelector, in the organization of the
	receiver, or answer nil if the receiver does not inlcude this selector."

	^organization categoryOfElement: aSelector! !

!AbstractProto methodsFor: 'accessing global variables (private)'!

globalVar: varName put: aValue
	"the global dictionary is shared by all prototypes of the same king,
	it is defined at the class level"
	^self class globalVar: varName put: aValue.!

globalVarValue: varName 
	"the global dictionary is shared by all prototypes of the same king,
	it is defined at the class level"
	^self class globalVarValue: varName! !

!AbstractProto methodsFor: 'subclass responsibility'!

hasVar: varName
	^false!

variableOwner: varName
	^nil! !

!AbstractProto methodsFor: 'workspace Interface'!

fileIn: aString
	(Filename named: aString) readStream fileInForProtoType: self class! !

!AbstractProto methodsFor: 'creating objects'!

addProperties: anArray
	self subclassResponsibility!

newEmpty
	^self class new!

newInitials: anArray
	"Has been replaced by the use of newEmpty or son followed by initialVars: and initialMethods: or initialSlots:"
	self error: 'See newInitials comment'! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

AbstractProto class
	instanceVariableNames: 'globalVarDictionary '!


!AbstractProto class methodsFor: 'creating new protos'!

new
	^super new initialize! !

!AbstractProto class methodsFor: 'initializing'!

initialize
	"AbstractProto initialize"
	UnknownMethod := Signal new notifierString: 'UnknownMethod: '.
	UnknownGlobalVariable := Signal new notifierString: 'UnknownGlobalVariable: '.
	globalVarDictionary := Dictionary new.
	ReturnFromSignal := Signal new.! !

!AbstractProto class methodsFor: 'exemple'!

exemple
	"create a workspace"
	self run
	"here are a few exemples of what you can type in the workspace"
	"
p1 := Root clone.
p1 method: 'x 10'.
p1 method: 'foo self x + self x'.
p1 method: 'bar self  x + self foo'.
p1 bar

p2 := Root clone.
p1 method: 'y p2'.
p3 := p1 clone.
p3 y.
p1 method: 'y 20'
p3 y.
p1 y."! !

!AbstractProto class methodsFor: 'browser interface'!

defaultBrowser
	^BasicProtoMethodBrowser! !

!AbstractProto class methodsFor: 'accessing'!

returnFromSignal
	^ReturnFromSignal!

unknownMethod
	^UnknownMethod! !

!AbstractProto class methodsFor: 'workspace Interface'!

workspace
	ProtoSpaceView openFor: self label: self printString , ' workspace'!

workspace: aText
	ProtoSpaceView openFor: self label: self printString , ' workspace' contents: aText! !

!AbstractProto class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
      aRoot addPrimitiveNamed: #browse.
      aRoot addPrimitiveNamed: #clone.
	aRoot addPrimitiveNamed: #printOn:.
	aRoot addPrimitiveNamed: #fileIn:.
	^aRoot!

createRoot
	| root |
	root := self new initialize.
	globalVarDictionary at: #Root put: root.
	^root!

run
	self run: ''!

run: aText
	self initialize.
	self addPrimitivesOn: self createRoot.
	self workspace: aText! !

!AbstractProto class methodsFor: 'smalltalk compiler interface'!

compile: aString 
	^Parser new
		parse: (ReadStream on: aString)
		class: nil
		noPattern: false
		context: nil
		notifying: (NonInteractiveCompilerErrorHandler new failBlock: [^nil])
		builder: self nodeBuilder new
		saveComments: true
		ifFail: [^nil]!

evaluate: aStream
	| methodNode context  |
	methodNode _ Parser new
				parse: aStream
				class: nil
				noPattern: true
				context: nil
				notifying: (InteractiveCompilerErrorHandler new editor: self; failBlock: [^#aborted])
				builder: self nodeBuilder new
				saveComments: true
				ifFail: [^nil].
	context := PContext new.
	context at: #self put: (globalVarDictionary at: #Root).
"	context at: #curRec put: (OrderedCollection with: (context at: #self))."
	^methodNode block body eval: context.!

evaluate: aStream for: anEditor
	| methodNode context  |
	methodNode _ Parser new
				parse: aStream
				class: nil
				noPattern: true
				context: nil
				notifying: (InteractiveCompilerErrorHandler new editor: anEditor; failBlock: [^#aborted])
				builder: self nodeBuilder new
				saveComments: true
				ifFail: [^nil].
	context := PContext new.
	context at: #self put: (globalVarDictionary at: #Root).
"	context at: #curRec put: (OrderedCollection with: (context at: #self))."
	^methodNode block body eval: context.!

nodeBuilder
	^ProgramNodeBuilder! !

!AbstractProto class methodsFor: 'accessing global variables (private)'!

globalVar: varName put: aValue
	"If the variable does not exists, it is created"
	globalVarDictionary at: varName put: aValue.
	^aValue!

globalVarValue: varName 
	"Access the global prototype dic and after the Smalltalk dic. this allow
	to access Smalltalk classes like Collections, etc"
	Dictionary keyNotFoundSignal
		handle: [:e1 | Dictionary keyNotFoundSignal
						handle: [:e2 | UnknownGlobalVariable raiseWith: varName]
						do: [^Smalltalk at: varName]]
		do: [^globalVarDictionary at: varName]! !

!AbstractProto class methodsFor: 'example'!

example
	self error: self printString , 'is an abstract class'.! !

!AbstractProto class methodsFor: 'file_out system'!

fileOutSystem
	"AbstractProto fileOutSystem"

	| fileStream |
	SystemOrganization fileOutCategory: #Prototypes.
	SystemOrganization fileOutCategory: #'Prototypes-ProgramNodes'.
	SystemOrganization fileOutCategory: #'Prototypes-Tools'.
	Object fileOutCategory:  #'prototype_simulation'.
	Behavior fileOutCategory:  #'prototype_simulation'.
	PeekableStream fileOutMessage:  #fileInForProtoType:.
	Metaclass fileOutProtocol: 'proto-evaluator'.

	fileStream := (Filename named: 'Metaclass_class-fileOutProtocol.st') writeStream.
	[fileStream timeStamp.
	Metaclass class fileOutCategory: 'fileOutProtocol'
		on: fileStream
		moveSource: false
		toFile: 0]
		valueNowOrOnUnwindDo: [fileStream close]

	"ProtoNeedingClasstalk must be filedOut manually"! !

AbstractProto subclass: #VMProto
	instanceVariableNames: 'variables methods '
	classVariableNames: 'UnknownVariable '
	poolDictionaries: ''
	category: 'Prototypes'!
VMProto comment:
'The variables are separated from methods.
They are stored in the instance variable: ''variables'' which is a dictionary.
Within a method, variables can represent either: arguments or temporaries, variables of the receiver object or global variables, in that order. If an argument of the current method is named ''x'' and if ''x'' is also a variable of the receiver, the former will hide the latter.
Variables cannot be accesed via message sending.
All the other characteristics of BasicProto are inherited.'!


!VMProto methodsFor: 'accessing variables (private)'!

hasVar: varName 
	^variables includesKey: varName!

hasVariables
	^variables isEmpty not!

localVar: varName put: aValue 
	variables at: varName put: aValue.
	^aValue!

localVarValue: varName 
	^variables at: varName!

var: varName put: aValue 
	^(self hasVar: varName) ifTrue: [self localVar: varName put: aValue] ifFalse: [self globalVar: varName put: aValue]!

varValue: varName 
	^(self hasVar: varName) ifTrue: [self localVarValue: varName] ifFalse: [self globalVarValue: varName]! !

!VMProto methodsFor: 'initializing (private)'!

initialize
	variables := Dictionary new.
	methods := Dictionary new.
	^super initialize.!

initMethDic: aDico
	methods := aDico!

initVarDic: aDico
	variables := aDico! !

!VMProto methodsFor: 'cloning (primitive)'!

basicClone
	| newDic newMethDic |
	newDic := Dictionary new.
	variables associationsDo: [:each | newDic add: each shallowCopy].
	newMethDic := Dictionary new.
	methods associationsDo: [:each | newMethDic add: each shallowCopy].
	^(self shallowCopy) initMethDic: newMethDic; initVarDic: newDic!

clone
	^self basicClone! !

!VMProto methodsFor: 'adding/removing methods (primitives)'!

addMethod: aString
	^self method: aString!

addMethod: aString classified: c
	^self method: aString classified: c!

addMethods: anArray
	anArray do: [:m | self addMethod: m]!

method: aString
	^self method: aString classified: 'as yet unclassified'!

method: aString classified: protocol 
	"interface method to add a new method to an object"
	"compile a string into a methodNode and store it the object"

	| methodNode selector |
	methodNode := self class compile: aString.
	selector := methodNode selector.
	methods at: selector put: methodNode.
	organization classify: selector under: protocol.!

removeMethod: selector
	""
	methods removeKey: selector.
	organization removeElement: selector.
	^selector! !

!VMProto methodsFor: 'accessing methods (private)'!

compiledMethodAt: name 
	^methods at: name!

hasMethods
	^methods isEmpty not!

includesSelector: name
	^(methods includesKey: name)!

methodNamed: name 
	^methods at: name ifAbsent: [nil]!

methodOwner: name
	"return the owner is the method is present and nil otherwise"
	^(self includesSelector: name) ifTrue: [self] ifFalse: [nil]! !

!VMProto methodsFor: 'adding primitive (private)'!

addPrimitive: aCompiledMethod named: sel
	"allow to define a primitive, i.e. a Smalltalk method in a proto dictionary, this method
	can be invoked like other methods by message sending"

	"store a Smalltalk method in the object"
	methods at: sel put: aCompiledMethod.
	organization classify: sel under: 'primitives'.
	^sel! !

!VMProto methodsFor: 'adding/removing variables (primitives)'!

addVar: varName
	"interface method to add a new slot to an object"
	^self addVar: varName value: nil.!

addVar: varName value: val
	"interface method to add a new slot to an object"
	^self internalAddVar: varName value: val!

addVariables: initArray 
	"adding several variables stored in initArray"
	"A first simple version that only deals with literals as variables values"

	| index size |
	index := 1.
	size := initArray size.
	[index < size]
		whileTrue: 
			[self addVar: (initArray at: index)
				value: (initArray at: index + 1).
			index := index + 2]!

internalAddVar: varName value: val
	"This should never be provided as aprimitive"
	Dictionary keyNotFoundSignal
		handle: [:e | variables at: varName put: val. ^self]
		do: [variables at: varName].
	self error: 'This variable already exists.'!

removeVar: selector
	""
	variables removeKey: selector.
	^selector! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

VMProto class
	instanceVariableNames: ''!


!VMProto class methodsFor: 'initializing'!

initialize
	super initialize.
	UnknownVariable := Signal new notifierString: 'UnknownVariable: '.! !

!VMProto class methodsFor: 'ceating new objects'!

newInitialVars: varArray initialMethods: methArray! !

!VMProto class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
  	super addPrimitivesOn: aRoot.
	^aRoot! !

!VMProto class methodsFor: 'exemple'!

exemple
	self run
"p1 := Root clone.
p1 var: #x value: 10.
p1 var: #y value: (Root clone var: z value: 5; method: 'z z').
p1 method: 'foo x + x'.
p1 method: 'bar x + self foo'.
p1 bar"! !

!VMProto class methodsFor: 'accessing'!

unknownVariable
	^UnknownVariable! !

!VMProto class methodsFor: 'compiler interface'!

nodeBuilder
	^EncapsulationPNB! !

VMProto subclass: #VMModifiable
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

VMModifiable class
	instanceVariableNames: ''!


!VMModifiable class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
 	super addPrimitivesOn: aRoot.
      aRoot addPrimitiveNamed: #addMethod:.
      aRoot addPrimitiveNamed: #addMethod:classified:.	
      aRoot addPrimitiveNamed: #addVar:value:.
      aRoot addPrimitiveNamed: #addVar:.
      aRoot addPrimitiveNamed: #addMethods:.
	^aRoot! !

!VMModifiable class methodsFor: 'example'!

example
	"self example"
	self run: 'p1 := Root clone.
p1 addVar: #x value: 10.
p1 addVar: #y value: (Root clone addVar: #z value: 5; addMethod: ''z z'').
p1 addMethod: ''foo x + x''.
p1 addMethod: ''bar x + self foo''.
p1 bar

p2 := Root clone.
p1 addMethod: ''y p2''.
p3 := p1 clone.
p3 y.
p1 addMethod: ''y 20''
p3 y.
p1 y.'! !

VMModifiable subclass: #L1
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

L1 class
	instanceVariableNames: ''!


!L1 class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
 	super addPrimitivesOn: aRoot.
      aRoot addPrimitiveNamed: #newEmpty.
	^aRoot! !

L1 subclass: #L1AndExplicitDelegation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!
L1AndExplicitDelegation comment:
'Variables + methods, modifiable, ability to explicitly delegates messages but not variable access.
Variables explicit delegation is not yet implemented.'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

L1AndExplicitDelegation class
	instanceVariableNames: ''!


!L1AndExplicitDelegation class methodsFor: 'compiler interface'!

nodeBuilder
	^ExplicitDelegationEncapsulationPNB! !

!L1AndExplicitDelegation class methodsFor: 'example'!

example

	"self example"

	self run: 'ed0 := Root clone.
ed0 addMethod:  ''m1 	x + y''.

ed1 := Root clone.
ed1 addVar: #x value: 10.
ed1 addVar: #y value: 20.
ed1 addMethod: ''y 		y * 2''.
ed1 addMethod:  ''m1		ed0 delegates: #m1 withArgs: nil''.
ed1 m1. 30

ed2 := Root clone.
ed2 addVar: #x value: 15.
ed2 addVar: #y value:  25.
ed2 addMethod: ''y 		ed1 delegates: #y withArgs: nil''.
ed2 addMethod: ''m1		ed1 delegates: #m1 withArgs: nil''.

ed2 m1 65'! !

L1 subclass: #L1AndImplicitDelegation
	instanceVariableNames: 'parent '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!


!L1AndImplicitDelegation methodsFor: 'creating new protos'!

clone
	| new |
	new := super clone.
	new initializeParent: parent.
	^new!

son
	"the basic method to create new objects"
	"creates a new object of which parent is self"
	^self species new initializeParent: self! !

!L1AndImplicitDelegation methodsFor: 'initializing (private)'!

initializeParent: anObject
	parent := anObject.! !

!L1AndImplicitDelegation methodsFor: 'accessing variables (private)'!

var: varName put: aValue 
	| owner |
	owner := self variableOwner: varName.
	^owner isNil
		ifTrue: [self globalVar: varName put: aValue]
		ifFalse: [owner localVar: varName put: aValue]!

variableOwner: varName
	^self varLookup: varName.!

varLookup: varName
	"return the owner of the slot or nil"
	Dictionary keyNotFoundSignal
		handle: [:e | parent isNil ifTrue: [^nil] ifFalse: [^parent varLookup: varName]]
		do: [variables at: varName. ^self]!

varValue: varName 
	| owner |
	owner := self variableOwner: varName.
	^owner isNil
		ifTrue: [self globalVarValue: varName]
		ifFalse: [owner localVarValue: varName]! !

!L1AndImplicitDelegation methodsFor: 'browserInterface'!

browseParent
	parent isNil ifTrue: [self error: 'top of the hierarchy'] ifFalse: [parent browse]! !

!L1AndImplicitDelegation methodsFor: 'to be removed'!

parent
	^parent! !

!L1AndImplicitDelegation methodsFor: 'accessing slots (private)'!

methodLookup: name
	"return the owner of the method or nil"
	Dictionary keyNotFoundSignal
		handle: [:e | parent isNil ifTrue: [^nil] ifFalse: [^parent methodLookup: name]]
		do: [methods at: name. ^self]!

methodOwner: name
	^self methodLookup: name! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

L1AndImplicitDelegation class
	instanceVariableNames: ''!


!L1AndImplicitDelegation class methodsFor: 'defaultBrowser'!

defaultBrowser
	^ImplicitDelegationMethodBrowser! !

!L1AndImplicitDelegation class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
	super addPrimitivesOn: aRoot.
	aRoot addPrimitiveNamed: #son.
	^aRoot! !

!L1AndImplicitDelegation class methodsFor: 'compiler interface'!

nodeBuilder
	^ImplicitDelegationEncapsulationPNB! !

!L1AndImplicitDelegation class methodsFor: 'example'!

example
	"Prototypes with variables and methods, creation ex-nihilo and 
	dynamic modification of structure, implicit delegation"
	"self example"

	self run: 'id1 := Root son.
id1 addVar: #x value: 10.
id1 addVar: #y value: 20.
id1 addMethod: ''foo x + y''.
id1 addMethod: ''bar: z z + self foo''.
id2 := id1 son.
id2 addVar: #x value: 30.
id2 addMethod: ''foo y''.
id3 := id1 son.
id3 addVar: #x value: 40.
id1 bar: 1. 31
id2 bar: 1. 21
id3 bar: 1. 61

turtle := Root son.
turtle addVar: #position value: 0@0.
turtle addVar: #direction value: #north.'! !

L1AndImplicitDelegation subclass: #ObjectLispLike
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!
ObjectLispLike comment:
'Object Lisp Simulation.

The fonction ask is implemented as a special form of the evaluator.'!


!ObjectLispLike methodsFor: 'classes simulation'!

exists: initList!

oneof! !

!ObjectLispLike methodsFor: 'defining functions'!

defobfun: aString
	^self method: aString! !

!ObjectLispLike methodsFor: 'adding variables to objects (primitive)'!

have: varName
	^self addVar: varName!

have: varName value: v
	^self addVar: varName value: v! !

!ObjectLispLike methodsFor: 'creating new objects (primitive)'!

kindof
	"creates a new object, the parent of which is the receiver"
	^self son! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

ObjectLispLike class
	instanceVariableNames: ''!


!ObjectLispLike class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
	super addPrimitivesOn: aRoot.
	aRoot removeMethod: #son.
      aRoot removeMethod: #addMethod:.
      aRoot removeMethod: #addMethod:classified:.	
      aRoot removeMethod: #addVar:value:.
      aRoot removeMethod: #addVar:.
      aRoot removeMethod: #addMethods:.
	aRoot addPrimitiveNamed: #kindof.
	aRoot addPrimitiveNamed: #have:.
	aRoot addPrimitiveNamed: #defobfun:.
	^aRoot! !

!ObjectLispLike class methodsFor: 'Compiler interface'!

nodeBuilder
	^ObjectLispLikePNB! !

!ObjectLispLike class methodsFor: 'example'!

example
	"self example"

	self run: '
fourthGrader := Root kindof.
fourthGrader have: #teacher.
fourthGrader ask: (teacher := #Joe).
fourthGrader ask: (teacher).
Global := 33.
fourthGrader ask: (Global).
Amanda := fourthGrader kindof.
Amanda ask: (teacher)
fourthGrader defobfun: ''say: sentence
	Transcript show: sentence; cr.''
Amanda say: ''Hello''.
Amanda ask: ( self say: ''hello'')
'! !

VMProto subclass: #VMNonModifiableL5
	instanceVariableNames: 'hasUserMethods '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!
VMNonModifiableL5 comment:
'Corresponds to the language L5 in the oopsla''92 paper.'!


!VMNonModifiableL5 methodsFor: 'adding initial methods'!

initialMethods: initArray 
	"Check that the receiver has not already methods"

	self hasUserMethods
		ifTrue: [self error: 'initialMethods can only be called once for non modifiable protos']
		ifFalse: 
			[self addMethods: initArray.
			hasUserMethods := true]! !

!VMNonModifiableL5 methodsFor: 'adding initial variables'!

initialVariables: initArray 
	"A first simple version that only deals with literals as variables values"

	|   |
	self hasVariables
		ifTrue: [self error: 'initialVariables can only be called once for non modifiable protos']
		ifFalse: 
			[self addVariables: initArray]! !

!VMNonModifiableL5 methodsFor: 'initializing'!

initialize
	super initialize.
	hasUserMethods := false! !

!VMNonModifiableL5 methodsFor: 'testing'!

hasUserMethods
	"There is here a problem due to the ways primitive are handled in that system
	as far as primitives are stored in objects, objets have methods before user defined
	ones are defined."
	"this solution must be changed"
	^hasUserMethods! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

VMNonModifiableL5 class
	instanceVariableNames: ''!


!VMNonModifiableL5 class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
 	super addPrimitivesOn: aRoot.
      aRoot addPrimitiveNamed: #initialMethods:.
      aRoot addPrimitiveNamed: #initialVariables:.
	^aRoot! !

!VMNonModifiableL5 class methodsFor: 'example'!

example
	"self example"

	self run: 'p1 := Root clone.
p1 initialVariables: #(i 1 j 2).
p1 initialMethods: #(
	''m1 i + j''
	''m2 i + self m1'')
'! !

VMNonModifiableL5 subclass: #L5AndImplicitDelegation
	instanceVariableNames: 'parent '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!
L5AndImplicitDelegation comment:
'Variables and methods,
no dynamic modification of structure,
creation ex-nihilo
creation primitive: newInitials
Implicit delegation'!


!L5AndImplicitDelegation methodsFor: 'creating new protos'!

clone
	| new |
	new := super clone.
	new initializeParent: parent.
	^new!

son
	"the basic method to create new objects"
	"creates a new object of which parent is self"
	^self species new initializeParent: self! !

!L5AndImplicitDelegation methodsFor: 'initializing (private)'!

initializeParent: anObject
	parent := anObject.! !

!L5AndImplicitDelegation methodsFor: 'accessing variables (private)'!

var: varName put: aValue 
	| owner |
	owner := self variableOwner: varName.
	^owner isNil
		ifTrue: [self globalVar: varName put: aValue]
		ifFalse: [owner localVar: varName put: aValue]!

variableOwner: varName
	^self varLookup: varName.!

varLookup: varName
	"return the owner of the slot or nil"
	Dictionary keyNotFoundSignal
		handle: [:e | parent isNil ifTrue: [^nil] ifFalse: [^parent varLookup: varName]]
		do: [variables at: varName. ^self]!

varValue: varName 
	| owner |
	owner := self variableOwner: varName.
	^owner isNil
		ifTrue: [self globalVarValue: varName]
		ifFalse: [owner localVarValue: varName]! !

!L5AndImplicitDelegation methodsFor: 'accessing methodes (private)'!

methodLookup: name
	"return the owner of the method or nil"
	Dictionary keyNotFoundSignal
		handle: [:e | parent isNil ifTrue: [^nil] ifFalse: [^parent methodLookup: name]]
		do: [methods at: name. ^self]!

methodOwner: name
	^self methodLookup: name! !

!L5AndImplicitDelegation methodsFor: 'browserInterface'!

browseParent
	parent isNil ifTrue: [self error: 'top of the hierarchy'] ifFalse: [parent browse]! !

!L5AndImplicitDelegation methodsFor: 'should be removed'!

parent
	^parent! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

L5AndImplicitDelegation class
	instanceVariableNames: ''!


!L5AndImplicitDelegation class methodsFor: 'defaultBrowser'!

defaultBrowser
	^ImplicitDelegationMethodBrowser! !

!L5AndImplicitDelegation class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
	super addPrimitivesOn: aRoot.
	aRoot addPrimitiveNamed: #son.
	^aRoot! !

!L5AndImplicitDelegation class methodsFor: 'exemple'!

exemple
	"create a workspace"
	self run
	"type expressions like the following:
IDPen1 := Root son.
IDPen1 addVar: #x value: 10.
IDPen1 addVar: #y value: 20.
IDPen1 addMethod: 'foo x + y'.
IDPen1 addMethod: 'bar: z z + self foo'.
IDPen1 addMethod: 'setX: arg x := arg'.
IDPen1 browse

IDPen2 := IDPen1 son.
IDPen2 addVar: #x value: 30.
IDPen2 addMethod: 'foo y'.
IDPen2 addMethod: 'setY: arg y := arg'.
IDPen2 addMethod: 'truc IDPen3 bar: 1'.

IDPen3 := IDPen1 son.
IDPen3 addVar: #x value: 40.

IDPen1 foo 30
IDPen2 foo 20
IDPen3 foo  60

IDPen1 bar: 1 31
IDPen2 bar: 1 21
IDPen3 bar: 1 61

IDPen2 truc 61

IDPen4 := IDPen3 clone.
IDPen4 bar: 1

IDPen3 setX: 15.
IDPen3 foo
"! !

!L5AndImplicitDelegation class methodsFor: 'compiler interface'!

nodeBuilder
	^ImplicitDelegationEncapsulationPNB! !

!L5AndImplicitDelegation class methodsFor: 'example'!

example
	"self example"

	self run: 'IDPen1 := Root son.
IDPen1 initialVariables: #(x 10 y 20).

IDPen1 initialMethods: #(
	''foo x + y'' 
	''bar: z z + self foo''
	''setX: arg x := arg''
).
IDPen1 bar: 2.

IDpen2 := IDPen1 son.
IDpen2 initialVariables: #(x 30).
IDpen2 bar: 2'! !

L5AndImplicitDelegation subclass: #L13
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

L13 class
	instanceVariableNames: ''!


!L13 class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
  	super addPrimitivesOn: aRoot.
      aRoot addPrimitiveNamed: #clone.
	^aRoot! !

L5AndImplicitDelegation subclass: #Act1Like
	instanceVariableNames: 'proxies '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!
Act1Like comment:
'Not in a coherent state.'!


!Act1Like methodsFor: 'creating new protos (primitive)'!

clone
	"create a new proto with the same variables, methodes and proxies than the modele.
	the methods are the sames but not the dictionaries containing them"
	^super clone initProxies: proxies shallowCopy.!

cloneWithProxies: aCollection
	"create a new proto with the same variables, methodes and proxies than the model.
	the methods are the sames but not the dictionaries containing them"
	^super clone initProxies: aCollection.!

cloneWithProxy: aProto
	"create a new proto with the same variables, methodes and proxies than the model.
	the methods are the sames but not the dictionaries containing them"
	^super clone initProxies: (OrderedCollection with: aProto).! !

!Act1Like methodsFor: 'testing (private)'!

canDelegatesTo: rec
	^proxies includes: rec! !

!Act1Like methodsFor: 'initializing (private)'!

initialize
	^super initialize initProxies: OrderedCollection new.!

initProxies: aCollection
	proxies := aCollection.! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Act1Like class
	instanceVariableNames: ''!


!Act1Like class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
	super addPrimitivesOn: aRoot.
	aRoot addPrimitiveNamed: #cloneWithProxy:.
	aRoot addPrimitiveNamed: #cloneWithProxies:.
	^aRoot! !

L5AndImplicitDelegation subclass: #L14
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!
L14 comment:
'Here we should handle split-objects'!


VMNonModifiableL5 subclass: #L15
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!
L15 comment:
'VMProto, non modifiable, explicit delegation'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

L15 class
	instanceVariableNames: ''!


!L15 class methodsFor: 'compiler interface'!

nodeBuilder
	^ExplicitDelegationEncapsulationPNB! !

VMProto subclass: #L5
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

L5 class
	instanceVariableNames: ''!


!L5 class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
 	super addPrimitivesOn: aRoot.
      aRoot addPrimitiveNamed: #newInitials:.
	^aRoot! !

AbstractProto subclass: #SProto
	instanceVariableNames: 'slots '
	classVariableNames: 'UnknownVariable '
	poolDictionaries: ''
	category: 'Prototypes'!


!SProto methodsFor: 'adding/removing slots (primitives)'!

addSlot: aString
	^self method: aString classified: 'as yet unclassified'!

method: aString
	^self method: aString classified: 'as yet unclassified'!

method: aString classified: protocol 
	"interface method to add a new method to an object"
	"compile a string into a methodNode and store it the object"

	| methodNode selector |
	methodNode := self class compile: aString.
	selector := methodNode selector.
	slots at: selector put: methodNode.
	organization classify: selector under: protocol.
	^selector!

removeMethod: selector
	""
	slots removeKey: selector.
	organization removeElement: selector.
	^selector! !

!SProto methodsFor: 'accessing slots (private)'!

compiledMethodAt: name 
	^slots at: name!

includesSelector: name
	^(slots includesKey: name)!

methodNamed: name 
	^slots at: name ifAbsent: [nil]!

methodOwner: name
	"return the owner is the method is present and nil otherwise"
	^(self includesSelector: name) ifTrue: [self] ifFalse: [nil]! !

!SProto methodsFor: 'initializing (private)'!

initialize
	slots := Dictionary new.
	^super initialize!

initMethDic: aDico
	slots := aDico! !

!SProto methodsFor: 'cloning (primitive)'!

clone
	| newDic |
	newDic := Dictionary new.
	slots associationsDo: [:each | newDic add: each copy].
	^self shallowCopy initMethDic: newDic! !

!SProto methodsFor: 'adding primitive (private)'!

addPrimitive: aCompiledMethod named: sel
	"allow to define a primitive, i.e. a Smalltalk method in a proto dictionary, this method
	can be invoked like other methods by message sending"

	"store a Smalltalk method in the object"
	slots at: sel put: aCompiledMethod.
	organization classify: sel under: 'primitives'.
	^sel! !

SProto subclass: #L11
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!


!L11 methodsFor: 'creatingObjects'!

initInitials: anArray
	anArray do: [:i | self addSlot: i]!

newInitials: anArray
	| x |
	x := self newEmpty.
	x initInitials: anArray.
	^x! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

L11 class
	instanceVariableNames: ''!


!L11 class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
 	super addPrimitivesOn: aRoot.
      aRoot addPrimitiveNamed: #newInitials:.
	^aRoot! !

L11 subclass: #L11AndImplicitDelegation
	instanceVariableNames: 'parent '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!


!L11AndImplicitDelegation methodsFor: 'creating new protos'!

clone
	| new |
	new := super clone.
	new initializeParent: parent.
	^new!

son
	"the basic method to create new objects"
	"creates a new object of which parent is self"
	^self species new initializeParent: self! !

!L11AndImplicitDelegation methodsFor: 'initializing (private)'!

initializeParent: anObject
	parent := anObject.! !

!L11AndImplicitDelegation methodsFor: 'accessing slots (private)'!

methodLookup: name
	"return the owner of the method or nil"
	Dictionary keyNotFoundSignal
		handle: [:e | parent isNil ifTrue: [^nil] ifFalse: [^parent methodLookup: name]]
		do: [slots at: name. ^self]!

methodOwner: name
	^self methodLookup: name! !

!L11AndImplicitDelegation methodsFor: 'browserInterface'!

browseParent
	parent isNil ifTrue: [self error: 'top of the hierarchy'] ifFalse: [parent browse]! !

!L11AndImplicitDelegation methodsFor: 'to be removed'!

parent
	^parent! !

!L11AndImplicitDelegation methodsFor: 'creatingObjects'!

sonWithInitialSlots: anArray
	| x |
	x := self son.
	x initInitials: anArray.
	^x! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

L11AndImplicitDelegation class
	instanceVariableNames: ''!


!L11AndImplicitDelegation class methodsFor: 'defaultBrowser'!

defaultBrowser
	^ImplicitDelegationMethodBrowser! !

!L11AndImplicitDelegation class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
	super addPrimitivesOn: aRoot.
	aRoot addPrimitiveNamed: #son.
	aRoot addPrimitiveNamed: #sonWithInitialSlots:.
	^aRoot! !

!L11AndImplicitDelegation class methodsFor: 'exemple'!

exemple
	self run
""! !

!L11AndImplicitDelegation class methodsFor: 'compiler interface'!

nodeBuilder
	^ImplicitDelegationPNB! !

!L11AndImplicitDelegation class methodsFor: 'example'!

example
	self run
""! !

L11AndImplicitDelegation subclass: #L16
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

L16 class
	instanceVariableNames: ''!


!L16 class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
  	super addPrimitivesOn: aRoot.
      aRoot addPrimitiveNamed: #clone.
	^aRoot! !

SProto subclass: #SModifiable
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SModifiable class
	instanceVariableNames: ''!


!SModifiable class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
 	super addPrimitivesOn: aRoot.
      aRoot addPrimitiveNamed: #addSlot:.
	^aRoot! !

SModifiable subclass: #L7
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

L7 class
	instanceVariableNames: ''!


!L7 class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
 	super addPrimitivesOn: aRoot.
      aRoot addPrimitiveNamed: #newEmpty.
	^aRoot! !

L7 subclass: #L7AndImplicitDelegation
	instanceVariableNames: 'parent '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!


!L7AndImplicitDelegation methodsFor: 'creating new protos'!

clone
	| new |
	new := super clone.
	new initializeParent: parent.
	^new!

son
	"the basic method to create new objects"
	"creates a new object of which parent is self"
	^self species new initializeParent: self! !

!L7AndImplicitDelegation methodsFor: 'initializing (private)'!

initializeParent: anObject
	parent := anObject.! !

!L7AndImplicitDelegation methodsFor: 'accessing slots (private)'!

methodLookup: name
	"return the owner of the method or nil"
	Dictionary keyNotFoundSignal
		handle: [:e | parent isNil ifTrue: [^nil] ifFalse: [^parent methodLookup: name]]
		do: [slots at: name. ^self]!

methodOwner: name
	^self methodLookup: name! !

!L7AndImplicitDelegation methodsFor: 'browserInterface'!

browseParent
	parent isNil ifTrue: [self error: 'top of the hierarchy'] ifFalse: [parent browse]! !

!L7AndImplicitDelegation methodsFor: 'to be removed'!

parent
	^parent! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

L7AndImplicitDelegation class
	instanceVariableNames: ''!


!L7AndImplicitDelegation class methodsFor: 'defaultBrowser'!

defaultBrowser
	^ImplicitDelegationMethodBrowser! !

!L7AndImplicitDelegation class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
	super addPrimitivesOn: aRoot.
	aRoot addPrimitiveNamed: #son.
	^aRoot! !

!L7AndImplicitDelegation class methodsFor: 'exemple'!

exemple
	self run
"IDP1 := Root son.
IDP1 addSlot: 'x 10'.
IDP1 addSlot: 'y 20'.
IDP1 addSlot: 'foo self x + self y'.
IDP1 addSlot: 'bar: z z + self foo'.
IDP1 bar: 5

IDP2 := IDP1 son.
IDP2 addSlot: 'x 30'.
IDP2 bar: 5"! !

!L7AndImplicitDelegation class methodsFor: 'compiler interface'!

nodeBuilder
	^ImplicitDelegationPNB! !

!L7AndImplicitDelegation class methodsFor: 'example'!

example
	"self example"
	self run: 'IDP1 := Root son.
IDP1 addSlot: ''x 10''.
IDP1 addSlot: ''y 20''.
IDP1 addSlot: ''foo self x + self y''.
IDP1 addSlot: ''bar: z z + self foo''.
IDP1 bar: 5.

IDP1 addSlot: ''printOn: aStream
	aStream nextPutAll: ''''aL7 with ID, x: ''''.
	self x printOn: aStream.
	aStream nextPutAll: '''', y: ''''.
	self y printOn: aStream.''

IDP1

IDP2 := IDP1 son.
IDP2 addSlot: ''x 30''.
IDP2 bar: 5'! !

L7AndImplicitDelegation subclass: #SelfLike
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!
SelfLike comment:
'This is the simulation of the Self choices for prototype-based programming.
Object only have methods, automatic accessors are generated for slots representing variables.
Within each method, idenfiers used as variables are interpreted as a message sending to the implicit receiver self.'!


!SelfLike methodsFor: 'adding slots'!

addSlot: chose 
	| truc stringName |
	chose isSymbol
		ifTrue: 
			[self method: chose asString , ' nil'.
			self assignmentMethod: chose asString , ': arg \"internal stuff"' withCRs accessPrimName: chose]
		ifFalse: 
			[truc := self isVarDefinition: chose.
			truc isNil
				ifFalse: 
					[stringName := truc at: 1.
					self method: stringName , ' nil'.
					self assignmentMethod: stringName , ': arg \"internal stuff"' withCRs accessPrimName: stringName asSymbol.
					(self compiledMethodAt: (stringName , ':') asSymbol)
						applyWith: (Array with: (truc at: 2))
						in: ((PContext new) at: #self put: self; yourself)]
				ifTrue: [self method: chose]]!

isVarDefinition: aString
	| index temp varName |
Object subscriptOutOfBoundsSignal
	handle: [:e | self error: 'syntax error']
	do: [index := 1.
		[(aString at: index) = $ ] whileTrue: [index := index + 1].
		temp := index.
		[(aString at: index) ~= $ ] whileTrue: [index := index + 1].
		varName := (aString copyFrom: temp to: (index - 1)) asSymbol.
		[(aString at: index) = $ ] whileTrue: [index := index + 1].
		((aString at: index) = $=) ifFalse: [^nil] ifTrue: [index := index + 1].
		[(aString at: index) = $ ] whileTrue: [index := index + 1].
		^Array with: varName with: (Compiler evaluate: (aString copyFrom: index to: aString size))]
		"temp := index.
		[index <= aString size and: [(aString at: index) ~= $ ]] whileTrue: [index := index + 1].
		^Array with: varName with: (Compiler evaluate: (aString copyFrom: temp to: (index - 1)))]"! !

!SelfLike methodsFor: 'adding assignment methods (private)'!

assignmentMethod: aString accessPrimName: n
	^self assignmentMethod: aString classified: 'as yet unclassified' accessPrimName: n!

assignmentMethod: aString classified: protocol accessPrimName: n
	"interface method to add a new method to an object"
	"compile a string into a methodNode and store it the object"

	| methodNode selector |
	methodNode _ Parser new
				parse: (ReadStream on: aString)
				class: nil
				noPattern: false
				context: nil
				notifying: (CompilerErrorHandler new failBlock: [^nil])
				builder: (SelfLikeAssignmentMethodBuilder new accessPrimName: n)
				saveComments: true
				ifFail: [^nil].
	selector := methodNode selector.
	slots at: selector put: methodNode.
	organization classify: selector under: protocol.
	^selector! !
"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

SelfLike class
	instanceVariableNames: ''!


!SelfLike class methodsFor: 'example'!

example
	"create a selfLike workspace"
	"self example"
	self run: '
sl1 := Root son.
sl1 addSlot: ''x = 5''.
sl1 x. 5
sl1 x: 10.
sl1 x. 10
sl1 addSlot: ''foo	x + x''.
sl1 addSlot: ''bar	x + foo''.
sl1 addSlot: ''baz: r
				| t |
				t := 1.
				r + t + foo''.
sl1 foo. 20
sl1 bar. 30
sl1 baz: 2. 23

sl2 := sl1 son.
sl2 x. 10
sl2 addSlot: ''x = 50''.
sl2 bar. 150

sl3 := sl1 clone.
sl3 bar. 30
'!

exemple
	"create a selfLike workspace"
	self run
	"in that workspace, you can type things like that:
sl1 := Root son.
sl1 addSlot: 'x = 5'.
sl1 x. 5
sl1 x: 10.
sl1 x. 10
sl1 addSlot: 'foo	x + x'.
sl1 addSlot: 'bar	x + foo'.
sl1 addSlot: 'baz: r
				| t |
				t := 1.
				r + t + foo'.
sl1 foo. 20
sl1 bar. 30
sl1 baz: 2. 23

sl2 := sl1 son.
sl2 x. 10
sl2 addSlot: 'x = 50'.
sl2 bar. 150

sl3 := sl1 clone.
sl3 bar. 30
"!

foo
"sl1 := Root son.
sl1 addSlot: 'x = 5'.
sl1 x. 5
sl1 x: 10.
sl1 x. 10
sl1 addSlot: 'foo	x + x'.
sl1 addSlot: 'bar	x + foo'.
sl1 addSlot: 'baz: r
				| t |
				t := 1.
				r + t + foo'.
sl1 foo. 20
sl1 bar. 30
sl1 baz: 2. 23

sl2 := sl1 son.
sl2 x. 10
sl2 addSlot: 'x = 50'.
sl2 bar. 150

sl3 := sl1 clone.
sl3 bar. 30
sl3 bar
"! !

!SelfLike class methodsFor: 'bootStrap'!

addPrimitivesOn: aRoot
	super addPrimitivesOn: aRoot.
      aRoot addPrimitiveNamed: #addSlot:.
	^aRoot! !

!SelfLike class methodsFor: 'smalltalk compiler interface'!

nodeBuilder
	^SelfLikePNB! !

L7 subclass: #L7AndExplicitDelegation
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Prototypes'!
L7AndExplicitDelegation comment:
'Slots + Modifiable + Explicit delegation.
As far as there is no difference between variables and methods, both are accessed via message sending,
so delegation works for both (see L1and ExplicitDelegation for a comparison).'!

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

L7AndExplicitDelegation class
	instanceVariableNames: ''!


!L7AndExplicitDelegation class methodsFor: 'exemple'!

exemple
	self run
"ed0 := Root clone.
ed0 addSlot:  'm1 	self x + self y'.

ed1 := Root clone.
ed1 addSlot: 'x 10'.
ed1 addSlot: 'y 20'.
ed1 addSlot:  'm1		ed0 delegates: #m1 withArgs: nil'.
ed1 m1. 30

ed2 := Root clone.
ed2 addSlot: 'x self z'.
ed2 addSlot: 'z 30'.
ed2 addSlot: 'y 		ed1 delegates: #y withArgs: nil'.
ed2 addSlot: 'm1		ed1 delegates: #m1 withArgs: nil'.

ed2 m1 50"! !

!L7AndExplicitDelegation class methodsFor: 'compiler interface'!

nodeBuilder
	^ExplicitDelegationPNB! !

!L7AndExplicitDelegation class methodsFor: 'example'!

example
	"self example"

	self run: 'ed0 := Root clone.
ed0 addSlot:  ''m1 	self x + self y''.

ed1 := Root clone.
ed1 addSlot: ''x 10''.
ed1 addSlot: ''y 20''.
ed1 addSlot:  ''m1		ed0 delegates: #m1 withArgs: nil''.
ed1 m1. 30

ed2 := Root clone.
ed2 addSlot: ''x self z''.
ed2 addSlot: ''z 30''.
ed2 addSlot: ''y 		ed1 delegates: #y withArgs: nil''.
ed2 addSlot: ''m1		ed1 delegates: #m1 withArgs: nil''.

ed2 m1 50'! !
AbstractProto initialize!
VMProto initialize!
